import types
from typing import Any, Dict, List, Optional, Union
from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.template import Template, TemplateFunction, TemplateRenderer
from abstra_internals.widgets.widget_base import Input


class ListInput:


    class LegacyListInput(Input):
        type = 'list-input'
        schemas: List[Any]
        item_schema: Any
        empty_value = []

        def __init__(self, key: str, item_schema: Any, **kwargs):
            initial_error = kwargs.get('errors', [])
            if not isinstance(initial_error, list):
                initial_error = [initial_error]
            super().__init__(key)
            self.item_schema = item_schema
            self.schemas = []
            self.value = self.empty_value
            self.set_props(dict(item_schema=item_schema, **kwargs))

        def serialize_value(self) ->List:
            return [schema.serialize_value() for schema in self.schemas]

        def set_props(self, props):
            self.min = props.get('min', 0)
            self.max = props.get('max', None)
            self.hint = props.get('hint', None)
            self.add_button_text = props.get('add_button_text', '+')
            self.full_width = props.get('full_width', False)
            self.required = props.get('required', True)
            self.disabled = props.get('disabled', False)
            self.schemas = [self.item_schema.copy() for _ in range(self.min)]
            if props.get('initial_value', None):
                self.set_value(props.get('initial_value', None), set_errors
                    =False)

        def _validate_list_min_max(self) ->List[str]:
            if self.min is not None and len(self.value) < self.min:
                return ['i18n_error_min_list']
            if self.max is not None and len(self.value) > self.max:
                return ['i18n_error_max_list']
            return []

        def _get_inputs(self):
            inputs = []
            for schema in self.schemas:
                inputs.extend(schema.get_input_widgets())
            return inputs

        def _validate_each(self) ->List[str]:
            inputs = self._get_inputs()
            for input in inputs:
                input.set_errors()
            if self._has_errors():
                return ['i18n_error_invalid_list_item']
            return []

        def _has_errors(self) ->bool:
            inputs = self._get_inputs()
            for input in inputs:
                has_errors = len(input.validate()) > 0
                if has_errors:
                    return True
            return False

        @property
        def validators(self):
            return super().validators + [self._validate_list_min_max] + [self
                ._validate_each]

        def render(self, ctx: dict):
            schemas = [schema.render(ctx) for schema in self.schemas]
            return {'type': self.type, 'key': self.key, 'hint': self.hint,
                'errors': self.errors, 'min': self.min, 'max': self.max,
                'addButtonText': self.add_button_text, 'fullWidth': self.
                full_width, 'required': self.required, 'disabled': self.
                disabled, 'schemas': schemas, 'value': self.serialize_value()}

        def set_value(self, value: List, set_errors=False):
            if not self._has_errors():
                self.errors = []
            if not value:
                value = []
            self.value = value
            for idx, value_item in enumerate(value):
                if idx >= len(self.schemas):
                    schema = self.item_schema.copy()
                    self.schemas.append(schema)
                else:
                    schema = self.schemas[idx]
                schema.set_values(value_item)
            self.schemas = self.schemas[:len(value)]

        def parse_value(self, value) ->List:
            parsed_values = []
            for index, value in enumerate(value or self.empty_value):
                if index >= len(self.schemas):
                    schema = self.item_schema.copy()
                    value = schema.parse_value(value or {})
                else:
                    schema = self.schemas[index]
                    value = schema.parse_value(value or {})
                parsed_values.append(value)
            return parsed_values


    class _ListInput(Input):
        type = 'list-input'
        _raw_template: Union[Template, TemplateFunction]
        empty_value = []
        value: List[State]

        def __init__(self, key: str, raw_template: Union[Template,
            TemplateFunction], **kwargs):
            initial_error = kwargs.get('errors', [])
            if not isinstance(initial_error, list):
                initial_error = [initial_error]
            super().__init__(key)
            self.value = self.empty_value
            self._raw_template = raw_template
            self.set_props(dict(**kwargs))

        def serialize_value(self):
            raise NotImplementedError(
                'Unnecessary as its done by rendering widgets')

        def get_item_template(self, idx: int) ->Template:
            item_state = self.value[idx] if idx < len(self.value) else {}
            if callable(self._raw_template):
                result = self._raw_template(State(item_state))
                if isinstance(result, tuple):
                    return result[0]
                return result
            return self._raw_template

        def set_props(self, props):
            self.min = props.get('min', 0)
            self.max = props.get('max', None)
            self.hint = props.get('hint', None)
            self.add_button_text = props.get('add_button_text', '+')
            self.full_width = props.get('full_width', False)
            self.required = props.get('required', True)
            self.disabled = props.get('disabled', False)
            super().set_props(props)

        def _validate_list_min_max(self) ->List[str]:
            if self.min is not None and len(self.value) < self.min:
                return ['i18n_error_min_list']
            if self.max is not None and len(self.value) > self.max:
                return ['i18n_error_max_list']
            return []

        def _validate_all_items(self) ->List[str]:
            for idx in range(len(self.value)):
                template = self.get_item_template(idx)
                renderer = TemplateRenderer(template)
                output = renderer.render(self.value[idx])
                if output['has_errors']:
                    return ['i18n_error_invalid_list_item']
            return []

        @property
        def validators(self):
            return super().validators + [self._validate_list_min_max, self.
                _validate_all_items]

        def render(self, ctx: dict):
            rendered_items = []
            serialized_value = []
            for idx in range(len(self.value)):
                item_template = self.get_item_template(idx)
                renderer = TemplateRenderer(item_template)
                output = renderer.render(self.value[idx])
                rendered_items.append(output['widgets'])
                item_value = {widget['key']: widget['value'] for widget in
                    output['widgets'] if 'key' in widget}
                serialized_value.append(item_value)
            return {'type': self.type, 'key': self.key, 'hint': self.hint,
                'errors': self.errors, 'min': self.min, 'max': self.max,
                'addButtonText': self.add_button_text, 'fullWidth': self.
                full_width, 'required': self.required, 'disabled': self.
                disabled, 'schemas': rendered_items, 'value': serialized_value}

        def parse_value(self, value: List[Optional[Dict]]) ->List[State]:
            parsed_values: List[State] = []
            if not value:
                return self.empty_value
            for idx, item in enumerate(value):
                state = State()
                template = self.get_item_template(idx)
                renderer = TemplateRenderer(template)
                raw_value = item or {}
                parsed = renderer.parse_state(raw_value)
                state.update_values(parsed)
                parsed_values.append(state)
            return parsed_values

    def __new__(cls, key: str, item_schema: Any, **kwargs):
        if isinstance(item_schema, list) or isinstance(item_schema, types.
            FunctionType):
            return cls._ListInput(key, raw_template=item_schema, **kwargs)
        return cls.LegacyListInput(key, item_schema, **kwargs)



import unittest
from typing import Dict, cast

from abstra_internals.entities.forms.form_entity import FormEntity
from abstra_internals.entities.forms.steps import (
    ComputationStep,
    GeneratorStep,
    PageStep,
    Step,
)
from abstra_internals.entities.forms.template import (
    BackButton,
    NextButton,
    State,
    Template,
    TemplateGeneratorFunction,
)
from abstra_internals.widgets.library import ListInput, TagInput, TextInput, TextOutput

rendered_text_input = {
    "type": "text-input",
    "key": "name",
    "label": "Name",
    "value": "",
    "placeholder": "",
    "required": True,
    "hint": None,
    "fullWidth": False,
    "mask": None,
    "maxLength": None,
    "minLength": None,
    "disabled": False,
    "errors": [],
}

rendered_text_output = {
    "type": "text-output",
    "text": "test",
    "size": "medium",
    "fullWidth": False,
}

rendered_list_input = {
    "type": "list-input",
    "key": "list",
    "hint": None,
    "errors": [],
    "min": 0,
    "max": None,
    "addButtonText": "+",
    "fullWidth": False,
    "required": True,
    "disabled": False,
    "schemas": [],
    "value": [],
}


def compare_renders(rendered, expected):
    assert rendered["end_page"] == expected["end_page"]
    assert rendered["steps_info"] == expected["steps_info"]
    assert len(rendered["widgets"]) == len(expected["widgets"])
    assert rendered["buttons"] == expected.get("buttons", [])
    for i, widget in enumerate(rendered["widgets"]):
        for key, value in widget.items():
            if value != expected["widgets"][i][key]:
                assert False, f"WIDGET '{widget['key']}' (index {i}) | KEY '{key}'\nEXPECTED: {expected['widgets'][i][key]}\nGOT: {value}"


def make_steps() -> Dict[str, Step]:
    def input_page(state: State) -> Template:
        return [TextInput(key="name", label="Name")]

    def double_input_page(state: State) -> Template:
        return [
            TextInput(key="name", label="Name"),
            TextInput(key="email", label="Email"),
        ]

    def output_page(state: State) -> Template:
        return [TextOutput(str(state["name"]))]

    def tag_input_page(state: State) -> Template:
        return [TagInput(key="tags", label="Tags")]

    def reactive_page(state: State) -> Template:
        p: Template = [TextInput(key="name", label="Name")]
        if state.get("name") == "test":
            p.append(TextOutput(str(state["name"])))
        return p

    def list_input_schema(item_state: State) -> Template:
        t: Template = [TextInput(key="name", label="Name")]
        if item_state.get("name") == "test":
            t.append(TextOutput(str(item_state["name"])))
            t.append(TextInput(key="email", label="Email"))
        return t

    def list_page(state: State) -> Template:
        return [ListInput(item_schema=list_input_schema, key="list")]

    def list_input_page(state: State):
        return [
            TextInput(key="name", label="Name"),
            ListInput(item_schema=list_input_schema, key="list"),
        ]

    def compute(state):
        pass

    return {
        "input_page": PageStep(input_page),
        "double_input_page": PageStep(double_input_page),
        "tag_input_page": PageStep(tag_input_page),
        "output_page": PageStep(output_page),
        "reactive_page": PageStep(reactive_page),
        "list_page": PageStep(list_page),
        "list_input_page": PageStep(list_input_page),
    }


class FormEntityTest(unittest.TestCase):
    def test_input_page(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "test"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

    def test_reactive_page(self):
        all_steps = make_steps()
        steps = [all_steps["reactive_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "test"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test",
                    },
                    {
                        **rendered_text_output,
                    },
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "tes"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "tes",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

    def test_validation(self):
        all_steps = make_steps()
        steps = [all_steps["double_input_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                    },
                    {
                        **rendered_text_input,
                        "key": "email",
                        "label": "Email",
                    },
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "my name"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "my name",
                    },
                    {
                        **rendered_text_input,
                        "key": "email",
                        "label": "Email",
                        "value": "",
                    },
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": ""},
            }
        )
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "",
                        "errors": ["i18n_error_required_field"],
                    },
                    {
                        **rendered_text_input,
                        "key": "email",
                        "label": "Email",
                        "value": "",
                    },
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

    def test_list_page(self):
        all_steps = make_steps()
        steps = [all_steps["list_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)

        # Initial render
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        # Add item
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"list": [None]},
            }
        )
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                        "schemas": [[rendered_text_input]],
                        "value": [{"name": ""}],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        # Fill item
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"list": [{"name": "test"}]},
            }
        )
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                        "schemas": [
                            [
                                {
                                    **rendered_text_input,
                                    "value": "test",
                                },
                                {
                                    **rendered_text_output,
                                    "text": "test",
                                },
                                {
                                    **rendered_text_input,
                                    "key": "email",
                                    "label": "Email",
                                    "value": "",
                                },
                            ]
                        ],
                        "value": [{"name": "test", "email": ""}],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

        # delete string
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"list": [{"name": ""}]},
            }
        )

        rendered = form.run()

        # validate error
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                        "schemas": [
                            [
                                {
                                    **rendered_text_input,
                                    "value": "",
                                    "errors": ["i18n_error_required_field"],
                                }
                            ]
                        ],
                        "errors": ["i18n_error_invalid_list_item"],
                        "value": [{"name": ""}],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
            },
        )

    def test_tag_input(self):
        all_steps = make_steps()
        steps = [all_steps["tag_input_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)

        # Initial render
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        "type": "tag-input",
                        "key": "tags",
                        "label": "Tags",
                        "value": [],
                        "placeholder": "",
                        "required": True,
                        "hint": None,
                        "fullWidth": False,
                        "disabled": False,
                        "errors": [],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
                "yielding": False,
            },
        )

        # Add item
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"tags": ["tag1"]},
            }
        )
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        "type": "tag-input",
                        "key": "tags",
                        "label": "Tags",
                        "value": ["tag1"],
                        "placeholder": "",
                        "required": True,
                        "hint": None,
                        "fullWidth": False,
                        "disabled": False,
                        "errors": [],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": True},
                "buttons": [NextButton()],
                "yielding": False,
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"tags": ["tag1", "tag2"]},
            }
        )

    def test_navigation(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"], all_steps["output_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [BackButton(), NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_back_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )

    def test_page_buttons(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"], all_steps["output_page"]]
        form = FormEntity(steps, State(), force_hide_steps=False)
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [BackButton(), NextButton()],
            },
        )

    def test_computation_step(self):
        def mock_compute(state):
            state["computed"] = "computed"

        all_steps = make_steps()
        steps = [
            all_steps["input_page"],
            ComputationStep(mock_compute),
            all_steps["output_page"],
        ]
        form = FormEntity(steps, State(), force_hide_steps=False)

        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [BackButton(), NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        self.assertEqual(form.state["computed"], "computed")

    def test_generator_step(self):
        def mock_generator(state: State):
            yield [TextOutput("test 1")]
            yield [TextOutput("test 2")]

        all_steps = make_steps()
        steps = [
            all_steps["input_page"],
            GeneratorStep(cast(TemplateGeneratorFunction, mock_generator)),
            all_steps["output_page"],
        ]
        form = FormEntity(steps, State(), force_hide_steps=False)

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test 1",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [],
            },
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test 2",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [],
            },
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
                "buttons": [BackButton(), NextButton()],
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_back_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
                "buttons": [NextButton()],
            },
        )
